# a minimal example
# %%
import heavylight

class ModelName(heavylight.Model):
    def t(self, t):
        return t
    
    # define user methods here
    def net_cashflow(self, t):
        """expected net cashflow occuring at time t"""
        if t == 0:
            return -100
        else:
            return self.monthly_premium(t) - 5
    
    def monthly_premium(self, t):
        return self.data['annual_premium'] / 12
    
    def forward_rate_pm(self, t):
        """monthly forward rate applying from time t-1 to time t (or however you define)"""
        return (1 + self.forward_rate_pa) ** (1/12) - 1
    
    def v(self, t):
        """discount factor from time t to time 0"""
        if t == 0:
            return 1.0
        else:
            return self.v(t - 1) / (1 + self.forward_rate_pm(t))
    
    def pv_net_cashflow(self, t):
        """present value (at t=0) of the cashflow occuring at time t"""
        return self.net_cashflow(t) * self.v(t)

# data for model.

data = {
    'annual_premium': 84,
}
# run model, we could store forward rate in another dictionary (like data), but for simplicity we can pass it in directly.
proj = ModelName(proj_len=5 * 12, data=data, forward_rate_pa=0.04)

# output the present value
print(f"Present Value of Net Cashflows: {proj.pv_net_cashflow.sum():=.2f}\n")

# view a dataframe of the results
print(proj.df.head())

# %%



# -*- coding: utf-8 -*-
# required libraries:
# heavylight
# pip install heavylight

# %%
import heavylight
from heavylight import Table
from protection_model_np import TermAssurance
import numpy as np
import pandas as pd

def solve_prot_premium(model: heavylight.Model, data: dict, basis: dict):
    
    data = data.copy()

    pricing_entries = {
            "annual_premium":1,
            "init_pols_if":1,
            }
    
    data.update(pricing_entries)

    proj_len = int(max(data["term_y"]) * 12 + 1)
    model_inst = model(data=data, basis=basis, proj_len=proj_len)
            
    # extract npvs
    npv_claims = model_inst.pv_claims.sum()
    npv_expenses = model_inst.pv_expenses.sum()
    npv_premiums = model_inst.pv_premiums.sum()
            
    # calculate premium
    
    annual_risk_premium = (npv_claims + npv_expenses) / npv_premiums
    monthly_risk_premium = annual_risk_premium / 12
    return np.round(monthly_risk_premium, 2)

# %%
if __name__=='__main__':
    
    basis = {
        "cost_inflation_pa": 0.02,
        "initial_expense": 500,
        "expense_pp": 10,
        "lapse_rate_pa": 0.1,
        "mort_table": Table.read_csv(r"tables/q_x_generic.csv"),
        "forward_rates": Table.read_csv(r"tables/forward_rates.csv"),
    }

    data = {
        "sum_assured": 100_000,
        "age_at_entry": 49,
        "term_y": 35,
        "smoker_status": "S",
        "shape": "level",
        "annual_premium": 1,
        "init_pols_if": 1,
        "extra_mortality": 0,
        "sex": "F"
    }
    data = {k:np.array([v] * 3) for k, v in data.items()}

    rng = np.random.default_rng(seed=42)

    # override the single datapoint `pols`
    pols = 10_000
    data = dict(
        sum_assured = rng.integers(10_000, 250_000, pols),
        age_at_entry = rng.integers(20, 50, pols),
        term_y = rng.integers(10, 30, pols),
        smoker_status = rng.choice(['S', 'N'], pols),
        shape = rng.choice(['level', 'decreasing'], pols),
        annual_premium = np.ones(pols),
        init_pols_if = np.ones(pols),
        extra_mortality = np.zeros(pols),
        sex = rng.choice(['F', 'M'], pols),
    )
    
    monthly_premium = solve_prot_premium(TermAssurance, data, basis)
    print("Premium: ", monthly_premium)

    do_output = input('Save output as result excel file? (Y/n):')
    if do_output[0].lower() == 'y':
        df = pd.DataFrame(data)
        df['monthly_premium'] = monthly_premium
        df.to_excel('output/projection_result.xlsx', index=False)
        print('Saved to output/projection_result.xlsx')
# %%
    



"""Testing the dataframe and cache presentation for LightModel class."""
from heavylight import LightModel, agg
from heavylight.memory_optimized_model import default_agg_function
import numpy as np
from typing import Callable, Union
import pandas as pd

class TestModel(LightModel):
    def __init__(self, agg_function: Union[Callable, None]):
        super().__init__(agg_function)
        self.size = 2

    def pols_if(self, t):
        if t == 0:
            return np.ones((self.size,))
        return self.pols_if(t - 1) - self.pols_death(t - 1)
    
    def pols_death(self, t):
        return .01 * self.pols_if(t)
    
    @agg(None)
    def test_agg_none(self, t):
        return np.ones((self.size,))
    
    @agg(len)
    def t(self, t):
        self.multi_param(t, t)
        return np.ones((self.size,))
    
    def multi_param(self, t, t2):
        return np.ones((self.size,))
    
expected_cache = {'pols_death': {0: np.array([0.01, 0.01]),
    1: np.array([0.0099, 0.0099])},
    'pols_if': {0: np.array([1., 1.]), 1: np.array([0.99, 0.99])},
    'test_agg_none': {0: np.array([1., 1.]), 1: np.array([1., 1.])},
    't': {0: np.array([1., 1.]), 1: np.array([1., 1.])},
    'multi_param': {(0, 0): np.array([1., 1.]), (1, 1): np.array([1., 1.])}}

expected_cache_no_multi = { k: v for k, v in expected_cache.items() if k != 'multi_param' }

expected_cache_agg = {'pols_if': {0: 2.0, 1: 1.98},
    'pols_death': {0: 0.02, 1: 0.0198},
    't': {0: 2, 1: 2},
    'multi_param': {(0, 0): 2, (1, 1): 2}}

expected_cache_agg_no_multi = { k: v for k, v in expected_cache_agg.items() if k != 'multi_param' }

expected_cache_agg_none_aggfunc = {'t': {0: 2, 1: 2}}

def test_model_df_before_run():
    tm = TestModel(default_agg_function)
    assert set(tm.df_agg.columns) == set(['pols_if', 'pols_death', 't'])
    assert set(tm.df.columns) == set(['pols_if', 'pols_death', 'test_agg_none', 't'])
    assert len(tm.df) == 0
    assert len(tm.df_agg) == 0

def test_method_df():
    tm = TestModel(default_agg_function)
    assert len(tm.t.cache) == 0
    assert tm.t.df.columns == ['t'] and len(tm.t.df) == 0
    tm.RunModel(1)
    assert len(tm.t.cache) == 2
    compare_structures(tm.t.cache, expected_cache['t'])
    assert tm.t.df.equals(pd.DataFrame({"t": tm.t.cache}))
    assert tm.t.df_agg.equals(pd.DataFrame({"t": tm.t.cache_agg}))

def test_model_cache_df_after_run():
    tm = TestModel(default_agg_function)
    tm.RunModel(1)
    compare_structures(tm.cache_agg, expected_cache_agg)
    compare_structures(tm.cache, expected_cache)
    assert tm.df.sort_index(axis=1).equals(pd.DataFrame(expected_cache_no_multi).sort_index(axis=1))
    assert tm.df_agg.sort_index(axis=1).equals(pd.DataFrame(expected_cache_agg_no_multi).sort_index(axis=1))

def test_model_cache_df_after_run_agg_none():
    tm = TestModel(None)
    assert set(tm.df.columns) == set(['pols_if', 'pols_death', 'test_agg_none', 't'])
    assert set(tm.df_agg.columns) == set(['t'])
    tm.RunModel(1)
    compare_structures(tm.cache_agg, expected_cache_agg_none_aggfunc)
    compare_structures(tm.cache, expected_cache)
    assert tm.df.sort_index(axis=1).equals(pd.DataFrame(expected_cache_no_multi).sort_index(axis=1))
    assert tm.df_agg.sort_index(axis=1).equals(pd.DataFrame(expected_cache_agg_none_aggfunc).sort_index(axis=1))

def test_model_t_first_column():
    tm = TestModel(default_agg_function)
    assert tm.df.columns[0] == 't'
    assert tm.df_agg.columns[0] == 't'
    tm.RunModel(1)
    assert tm.df.columns[0] == 't'
    assert tm.df_agg.columns[0] == 't'

def compare_structures(obj1, obj2):
    if isinstance(obj1, dict) and isinstance(obj2, dict):
        assert set(obj1.keys()) == set(obj2.keys()), "Dictionary keys do not match."
        for key in obj1:
            compare_structures(obj1[key], obj2[key])
    elif isinstance(obj1, np.ndarray) and isinstance(obj2, np.ndarray):
        assert np.array_equal(obj1, obj2), "Numpy arrays are not equal."
    else:
        assert obj1 == obj2, f"Values do not match: {obj1} != {obj2}"


import numpy as np
from heavylight import LightModel
import numpy as np
import pytest

class SimpleModel(LightModel):

    def __init__(self, initial_pols_if: np.ndarray, mortality_rate = .01):
        super().__init__()
        self.initial_pols_if = initial_pols_if
        self.mortality_rate = mortality_rate

    def t(self, t):
        return np.int64(t) # to calculate nbytes for memory consumption

    def num_pols_if(self, t):
        if t == 0:
            return self.initial_pols_if
        else:
            return self.num_pols_if(t - 1) - self.pols_death(t - 1) # causes exponential time complexity if uncached
        
    def pols_death(self, t):
        return self.num_pols_if(t) * self.mortality_rate

    def cashflow(self, t):
        return self.num_pols_if(t) * 100
    
    def v(self, t):
        if t == 0:
            return np.int64(1)
        else:
            return self.v(t - 1) / (1 + self.forward_rate(t))
    
    def forward_rate(self, t):
        return np.float64(0.04)
    
    def pv_cashflow(self, t):
        return self.cashflow(t) * self.v(t)

def calculate_cache_graph_size(model: LightModel):
    cg = model.cache_graph
    return sum(val.nbytes for cache in cg._caches.values() for val in cache.values())

def run_model_calculate_max_cache(model: SimpleModel, max_time: int):
    max_cache_size = 0
    for t in range(max_time+1):
        model.pv_cashflow(t)
        cache_size = calculate_cache_graph_size(model)
        max_cache_size = max(max_cache_size, cache_size)
    return max_cache_size

def get_memory_savings_ratio(model: SimpleModel, max_time: int):
    model.ResetCache()
    max_cache_size = run_model_calculate_max_cache(model, max_time)
    model.OptimizeMemoryAndReset()
    optimized_max_cache_graph_size = run_model_calculate_max_cache(model, max_time)
    return optimized_max_cache_graph_size / max_cache_size

@pytest.mark.timeout(4)
def test_memory_savings():
    model = SimpleModel(np.ones((1000,)))
    assert get_memory_savings_ratio(model, 100) < .01
    assert sum(model.cache_graph.cache_misses.values()) > 0 # verify that misses are being tracked
    assert all(misses == 1 for misses in model.cache_graph.cache_misses.values())

def test_cache_misses():
    sm = SimpleModel(np.linspace(.1, 1, 10))
    sm.RunModel(5)
    assert len(sm.cache_graph.cache_misses.values()) > 0
    assert all(x == 1 for x in sm.cache_graph.cache_misses.values())
    sm.OptimizeMemoryAndReset()
    assert len(sm.cache_graph.cache_misses.values()) == 0
    sm.RunModel(5)
    assert len(sm.cache_graph.cache_misses.values()) > 0
    assert all(x == 1 for x in sm.cache_graph.cache_misses.values())



import numpy as np
from heavylight import LightModel
from heavylight.memory_optimized_model import agg
from typing import Union, Callable
import pytest

class SimpleModel(LightModel):

    def __init__(self, initial_pols_if: np.ndarray, mortality_rate = .01):
        super().__init__()
        self.initial_pols_if = initial_pols_if
        self.mortality_rate = mortality_rate

    def t(self, t):
        return t
    
    def num_pols_if(self, t):
        if t == 0:
            return self.initial_pols_if
        else:
            return self.num_pols_if(t - 1) - self.pols_death(t - 1) # causes exponential time complexity if uncached
    
    def pols_death(self, t):
        return self.num_pols_if(t) * self.mortality_rate

    def cashflow(self, t):
        return self.num_pols_if(t) * 100
    
    def v(self, t):
        """discount factor for time t --> time 0"""
        if t == 0:
            return 1
        else:
            return self.v(t - 1) / (1 + self.forward_rate(t))
    
    @agg(lambda x: x * 2)
    def forward_rate(self, t):
        return 0.04
    
    def pv_cashflow(self, t):
        """present value of the cashflow occuring at time t"""
        return self.cashflow(t) * self.v(t)
    
    def fib(self, t, mult_factor):
        if t <= 1:
            return t
        return mult_factor * self.fib(t - 1, mult_factor) + self.fib(t - 2, mult_factor)

def test_method_call_and_cache_retrievals():
    sm = SimpleModel(np.linspace(.1, 1, 10))
    sm.forward_rate(0)
    assert sm.forward_rate.cache[0] == .04
    assert sm.cache_graph._caches['forward_rate'][((0,), frozenset())] == .04
    assert sm.cache_graph.cache['forward_rate'][0] == .04
    assert len(sm.cache_graph.cache) == 1
    assert len(sm.forward_rate.cache) == 1
    sm.forward_rate(5)
    assert len(sm.forward_rate.cache) == 2
    assert len(sm.cache_graph.cache) == 1

def test_timestep_functions():
    sm = SimpleModel(np.linspace(.1, 1, 10))
    timestep_functions = sm.timestep_functions
    expected_functions = ['t', 'num_pols_if', 'pols_death', 'cashflow', 'v', 'pv_cashflow', 'forward_rate']
    assert set(timestep_functions) == set(expected_functions) # no duplicates

@pytest.mark.timeout(4)
def test_caching_speedups():
    sm = SimpleModel(np.linspace(.1, 1, 10))
    assert len(sm.num_pols_if.cache) == 0
    sm.RunModel(200)
    assert len(sm.num_pols_if.cache) == 201
    # fib did not get called because it is not single parameter time function
    assert len(sm.fib.cache) == 0
    sm.fib(200, 1.1)
    # but it is still cached
    assert len(sm.fib.cache) == 201

def test_reset_cache():
    sm = SimpleModel(np.linspace(.1, 1, 10))
    sm.RunModel(5)
    assert round(np.sum(sm.pols_death.cache[0]), 10) == .055
    sm.mortality_rate = .02
    sm.RunModel(5)
    assert round(np.sum(sm.pols_death.cache[0]), 10) == .055
    sm.ResetCache()
    sm.RunModel(5)
    assert round(np.sum(sm.pols_death.cache[0]), 10) == .11

class TestPrettyCacheModel(LightModel):
    def __init__(self):
        super().__init__()
    def t(self, t):
        return self.wowee(t, 1) + self.zowee(t, x=1)
    def wowee(self, t, x):
        return 1
    def zowee(self, t, x):
        return 2
    
def test_pretty_cache():
    pretty_model = TestPrettyCacheModel()
    pretty_model.ResetCache()
    pretty_model.RunModel(0)
    assert pretty_model.cache_graph.cache['wowee'][(0,1)] == 1
    assert pretty_model.wowee.cache[(0,1)] == 1
    assert pretty_model.cache_graph.cache['zowee'][((0,),frozenset({'x': 1}.items()))] == 2
    assert pretty_model.zowee.cache[((0,),frozenset({'x': 1}.items()))] == 2
    assert pretty_model.cache_graph.cache['t'][0] == 3
    assert pretty_model.t.cache[0] == 3
    # can inject into the cache
    pretty_model.zowee[1] = 'hello cache'
    assert pretty_model.zowee.cache[1] == 'hello cache'



# %%
import heavylight
import pandas as pd
import numpy as np

# A Vectorised Actuarial Contingency Model

# This code is out of date/unmaintained but includes some examples of usage.

# %%
class Life(heavylight.Model):
    time_step = 1/12

    def t(self, t):
        return np.ones(self.data["data_count"]) * t
    
    def num_pols_if(self, t):
        if t == 0:
            return self.data["initial_policies"]
        else:
            return self.num_pols_if(t - 1) - self.num_deaths(t - 1)
    
    def num_deaths(self, t):
        if t == 0:
            return np.zeros(self.data["data_count"])
        else:
            return self.num_pols_if(t) * self.q_x_m(t)
    
    def age(self, t):
        """age at time t"""
        if t == 0:
            return self.data["initial_age"] # floating point
        else:
            return self.age(t - 1) + Life.time_step
    
    def age_rounded(self, t):
        return np.round(self.age(t)).astype(int)
    
    def q_x_m(self, t):
        """monthly mortality rate"""
        return self.q_x(t) ** (Life.time_step)
    
    def q_x(self, t):
        """annual mortality rate"""
        return self.basis["mortality"][self.age_rounded(t)]
    
    def premium(self, t):
        #print(self.data["premium"])
        #print(self.num_pols_if(t))
        return self.data["premium"] * self.num_pols_if(t)
    
    def claim(self, t):
        #print(f"claim: {len(self.num_deaths(t))=}")
        #print(f"claim: {len(self.data['sum_assured'])=}")
        return self.num_deaths(t) * self.data["sum_assured"]

    def net_cashflow(self, t):
        return self.premium(t) - self.claim(t)
    
    def mpnum(self, t):
        return self.data["mp_num"]


# %%
mortality = heavylight.Table.read_csv("sample_q_x_table.csv")

# %%
basis = {"mortality": mortality}

# %%


def make_random_data(num_pols):
    rng = np.random.default_rng(seed=42)
    return_data = {}
    return_data["data_count"] = num_pols
    return_data["initial_policies"] = np.ones(num_pols)
    return_data["mp_num"] = np.arange(num_pols)
    return_data["initial_age"] = rng.uniform(low=20.0, high=25.0, size=num_pols)
    return_data["premium"] = rng.beta(a=2, b=5, size=num_pols) * 100 + 50
    return_data["sum_assured"] = return_data["premium"] * 100 # rng.uniform(low=10.0, high=20, size=num_pols)
    return return_data

"""
data = {
    "initial_policies": np.ones(num_model_points),
    "initial_age" : np.array([32+1/12, 42, 25+7/12]),
    "premium": np.array([100.65, 500, 220.34]),
    "sum_assured": np.array([10000, 30000, 25000]),
    }
"""

# num_model_points = 3


# %%
data = make_random_data(1000)
model = Life(data=data, basis=basis, proj_len=10) #, do_run=True, proj_len = 5)



# %%
def expand(model, variable):
    """return a variable from the model as a dataframe
    rows = time
    columns = variables
    """
    temp_df = pd.DataFrame(getattr(model, variable).values).T
    temp_df.index.name = "t"
    return temp_df
# %%
def get_single_result(model, index):
    temp_df = model.ToDataFrame()
    # TODO: check type of each column
    # TODO: check that index is not out of bounds
    for column in temp_df:
        if isinstance(temp_df[column].iloc[0], np.ndarray):
            temp_df[column] = temp_df[column][index]
    return temp_df
    #return temp_df.applymap(lambda x: x[index])

# %%

def dfize(model):
    columns = {}
    for func in model._funcs:
        func_values = getattr(model,func).values
        if isinstance(func_values.values[0], np.ndarray):
            # multi-index
            temp_df = pd.DataFrame(func_values).T.stack()
            columns[func] = temp_df
        else:
            print(f"Still to deal with {func}")
    return pd.concat(columns, axis=1)


    # convert 

# %%
df = dfize(model)
df[df.mpnum==500].plot(x="t", y="net_cashflow")  # can filter on a model point
# %%
def npv(rate, values):
    values = np.atleast_2d(values)
    timestep_array = np.arange(0, values.shape[1])
    npv = (values / (1 + rate) ** timestep_array).sum(axis=1)
    try:
        # If size of array is one, return scalar
        return npv.item()
    except ValueError:
        # Otherwise, return entire array
        return npv
# %%
print("npv", npv(0.08, df[df.mpnum==500]["net_cashflow"]))
# %%
npvs_4pc = df.groupby("mpnum").agg({"net_cashflow": lambda x: npv(0.04, x)})
# %%



from inspect import getmembers, signature
from typing import Callable, Dict, List, Union, Any
from types import MethodType
from heavylight.memory_optimized_cache import CacheGraph, CacheMethod, FunctionCall
import pandas as pd
import numpy as np


def default_agg_function(x: Any):
    """Default aggregation function for storing results."""
    if isinstance(x, np.ndarray) and issubclass(x.dtype.type, np.number):
        return np.sum(x)
    return x


class LightModel:
    """Base class to subclass from for recursive actuarial models.

    Inheriting from this class causes functions starting with a lowercase letter to be cached.
    A cached function will only be called once for each unique set of arguments.
    Caching is necessary for recursive actuarial calculations to be efficient.

    This means that the function will only be called once for each unique set of arguments, important for recursive actuarial models.

    Parameters
    ----------
    agg_function: Callable[[int], Any], optional
        This function is applied to the results of methods starting with a lowercase letter.

    Class level methods:
        RunModel(proj_len):
        if the model has not been auto-run at initialisation, run it for projection length.

    methods/variables to avoid:
    methods/variables starting with an underscore `_` are treated as internal.  You may break functionality if you create your own.
    """

    def __init__(self, agg_function: Union[Callable, None] = default_agg_function):
        self.cache_graph = CacheGraph()
        self._single_param_timestep_funcs: List[CacheMethod] = []
        self._funcs: Dict[MethodType, CacheMethod] = {}
        # happens after setting up attributes
        for method_name, method in getmembers(self):
            if (
                not method_name[0].islower()
                or method_name.startswith("_")
                or not isinstance(method, MethodType)
            ):
                continue
            method_agg_function = getattr(method, "agg_function", None)
            if not hasattr(method, "agg_function"): # only provide class-level agg_function if method-level is not provided
                method_agg_function = agg_function
            cached_method = self.cache_graph(method_agg_function)(method)
            self._funcs[method] = cached_method
            is_single_param_t = check_single_parameter_name(method, "t")
            if is_single_param_t:
                self._single_param_timestep_funcs.append(cached_method)
            setattr(self, method_name, cached_method)

    def RunModel(self, proj_len: int):
        """
        Run the model for a projection length.
        All single parameter timestep functions will be run for each timestep.
        """
        for t in range(proj_len + 1):
            for func in self._single_param_timestep_funcs:
                # We avoid recalling any functions that have already been cached, resolves issue #15 lewisfogden/heavylight
                if (
                    not FunctionCall(func.func.__name__, (t,), frozenset())
                    in self.cache_graph.all_calls
                ):
                    func(t)

    def ResetCache(self):
        """Reset the cache, useful if you want to run the model again with different parameters."""
        self.cache_graph.reset()

    def OptimizeMemoryAndReset(self):
        """
        Calling this clears the cache, and causes the cache to be optimized.
        Optimizing the cache means that only needed results are stored, and intermediate results are cleared.
        """
        self.cache_graph.optimize_and_reset()

    @property
    def timestep_functions(self):
        """List the cached functions that have a single parameter `t`."""
        return [cache.func.__name__ for cache in self._single_param_timestep_funcs]

    @property
    def cache(self):
        return self.cache_graph.cache

    @property
    def cache_agg(self):
        return self.cache_graph.cache_agg

    @property
    def df(self):
        return self.ToDataFrame(is_agg=False)

    @property
    def df_agg(self):
        return self.ToDataFrame(is_agg=True)

    def ToDataFrame(self, param_name="t", is_agg=False):
        """Return a pandas dataframe of all single parameter timestep results."""
        get_method_cache = _get_method_cache_factory(is_agg)
        filtered_cache = {}
        for method, cache_method in self._funcs.items():
            if not check_single_parameter_name(method, param_name):
                continue
            if is_agg and cache_method.agg_func is None:
                continue
            filtered_cache[method.__name__] = get_method_cache(cache_method)
        df = pd.DataFrame(filtered_cache)
        # if t is in the dataframe, move it to first position
        if "t" in df.columns:
            df.insert(0, "t", df.pop("t"))
        return df


def check_single_parameter_name(func: Callable, name: str):
    sig = signature(func)
    return len(sig.parameters) == 1 and name in sig.parameters


def agg(agg_function: Union[Callable, None]):
    """
    Register the storage function on a function.
    Used for storing aggregated results before cache eviction to reduce memory consumption.
    """

    def decorator(func: Callable):
        func.agg_function = agg_function # type: ignore
        return func

    return decorator


def _get_method_cache_factory(is_agg: bool):
    if is_agg:
        accessor: Callable[[CacheMethod], Dict] = lambda method: method.cache_agg
    else:
        accessor: Callable[[CacheMethod], Dict] = lambda method: method.cache
    return accessor



# creates a folder with examples

'''
usage:
```python
import heavylight
heavylight.make_example('example', 'path/to/folder')

or using a simple CLI

'''
import shutil
from pathlib import Path
from typing import Union, Literal

examples = ['template', 'protection']
ExampleName = Literal['template', 'protection']

def make_example(download_to_path: Union[str, Path], example_name: ExampleName):
    if not example_name in examples:
        raise ValueError(f'example {example_name} not in available examples {examples}')
    example_path = Path(__file__).resolve().parent / 'examples' / example_name
    dest_path = Path(download_to_path).resolve() / example_name
    print(f'Creating {example_name} at {dest_path}')
    shutil.copytree(src=example_path, dst=dest_path)


import heavylight
import numpy as np

class TermAssurance(heavylight.Model):
    def t(self, t):
        return t

    def net_cf(self, t):
        return self.premiums(t) - self.claims(t) - self.expenses(t)
    
    def premium_pp(self, t):
        """monthly premium"""
        return self.data["annual_premium"] / 12
    
    def sum_assured(self, t):
        return np.where(
            self.data["shape"] == 'level',
            self.data["sum_assured"],
            self.decreasing_sa(t)
        )

    def decreasing_sa(self, t):
        """sum assured if decreasing, tracking mortgage - interest rate of 7% assumed (could be parameterised)"""
        r = (1 + 0.07)**(1/12) - 1
        S = self.data["sum_assured"]
        T = self.data["term_y"] * 12
        outstanding = S * ((1 + r)**T - (1 + r)**t)/((1 + r)**T - 1)
        return outstanding

    def claim_pp(self, t):
        return np.where(
            t > self.data["term_y"] * 12,
            0,
            self.sum_assured(t)
        )

    def inflation_factor(self, t):
        """annual"""
        return (1 + self.basis["cost_inflation_pa"])**(t // 12)

    def v(self, t):
        """present value of 1 discounted from time t to time 0"""
        if t == 0:
            return 1
        else:
            return self.v(t - 1) / (1 + self.basis["forward_rates"][t])

    def premiums(self, t):
        return self.premium_pp(t) * self.num_pols_if(t)
    
    def pv_premiums(self, t):
        return self.v(t) * self.premiums(t)
    
    def duration(self, t):
        """duration in force in years"""
        return t // 12
    
    def claims(self, t):
        return self.claim_pp(t) * self.num_deaths(t)
    
    def pv_claims(self, t):
        return self.v(t) * self.claims(t)
      
    def expenses(self, t):
        if t == 0:
            return self.basis["initial_expense"]
        else:
            return self.num_pols_if(t) * self.basis["expense_pp"] / 12 * self.inflation_factor(t)
      
    def pv_expenses(self, t):
        return self.v(t) * self.expenses(t)

    def num_pols_if(self, t):
        """number of policies in force"""
        if t == 0:
            return self.data["init_pols_if"]    
        # if statements need to be vectorised, using np.where (same logic as excel IF())
        else:
            return np.where(
                t > self.data["term_y"] * 12,
                0,
                self.num_pols_if(t - 1) - self.num_exits(t - 1) - self.num_deaths(t - 1)
            )
    
    def num_exits(self, t):
        """exits occurring at time t"""
        return self.num_pols_if(t) * (1 - (1 - self.basis["lapse_rate_pa"])**(1/12))
      
    def num_deaths(self, t):
        """deaths occurring at time t"""
        return self.num_pols_if(t) * self.q_x_12(t)
    
    def age(self, t):
        return self.data["age_at_entry"] + t // 12
    
    def q_x_12(self, t):
        return 1 - (1 - self.q_x_rated(t))**(1/12)
    
    def q_x(self, t):
        return self.basis["mort_table"][self.age(t), self.duration(t), self.data["smoker_status"]]
            
    def q_x_rated(self, t):
        return np.clip(self.q_x(t) * (1 + self.data["extra_mortality"]), 0, 1)
        
    def commission(self, t):
        return 0



import heavylight

class TermAssurance(heavylight.Model):
    def t(self, t):
        return t

    def net_cf(self, t):
        return self.premiums(t) - self.claims(t) - self.expenses(t)
    
    def premium_pp(self, t):
        """monthly premium"""
        return self.data["annual_premium"] / 12
    
    def claim_pp(self, t):
        if t == 0:
            return self.data["sum_assured"]
        elif t > self.data["term_y"] * 12:
            return 0
        elif self.data["shape"] == "level":
            return self.data["sum_assured"]
        elif self.data["shape"] == "decreasing":
            r = (1 + 0.07)**(1/12)-1
            S = self.data["sum_assured"]
            T = self.data["term_y"] * 12
            outstanding = S * ((1 + r)**T - (1 + r)**t)/((1 + r)**T - 1)
            return outstanding
        else:
            raise ValueError("Parameter 'shape' must be 'level' or 'decreasing'")
    
    def inflation_factor(self, t):
        """annual"""
        return (1 + self.basis["cost_inflation_pa"])**(t//12)

    def v(self, t):
        """present value of 1 discounted from time t to time 0"""
        if t == 0:
            return 1
        else:
            return self.v(t - 1) / (1 +self.basis["forward_rates"][t])

    def premiums(self, t):
        return self.premium_pp(t) * self.num_pols_if(t)
    
    def pv_premiums(self, t):
        return self.v(t) * self.premiums(t)
    
    def duration(self, t):
        """duration in force in years"""
        return t//12
    
    def claims(self, t):
        return self.claim_pp(t) * self.num_deaths(t)
    
    def pv_claims(self, t):
        return self.v(t) * self.claims(t)
      
    def expenses(self, t):
        if t == 0:
            return self.basis["initial_expense"]
        else:
            return self.num_pols_if(t) * self.basis["expense_pp"]/12 * self.inflation_factor(t)
      
    def pv_expenses(self, t):
        return self.v(t) * self.expenses(t)

    def num_pols_if(self, t):
        """number of policies in force"""
        if t == 0:
            return self.data["init_pols_if"]
        elif t > self.data["term_y"] * 12:
            return 0
        else:
            return self.num_pols_if(t-1) - self.num_exits(t-1) - self.num_deaths(t-1)
    
    def num_exits(self, t):
        """exits occurring at time t"""
        return self.num_pols_if(t) * (1-(1 - self.basis["lapse_rate_pa"])**(1/12))
      
    def num_deaths(self, t):
        """deaths occurring at time t"""
        return self.num_pols_if(t) * self.q_x_12(t)
    
    def age(self, t):
        return self.data["age_at_entry"] + t//12
    
    
    def q_x_12(self, t):
        return 1-(1- self.q_x_rated(t))**(1/12)
    
    def q_x(self, t):
        return self.basis["mort_table"][self.age(t), self.duration(t), self.data["smoker_status"]]
            
    def q_x_rated(self, t):
        return max(0, min(1 , self.q_x(t) * (1 + self.data["extra_mortality"]) ) )
        
    def commission(self, t):
        return 0



from heavylight import make_example
from pathlib import Path
import shutil
import pytest

def test_make_example_bad():
    with pytest.raises(ValueError):
        make_example('temp', 'bad')

def test_make_example_protection():
    make_example('temp', 'protection')
    assert (Path('temp') / 'protection').exists()
    shutil.rmtree('temp/protection')




# Heavytables
# Provides a high performance multiple-index -> single index table

import itertools
import pandas as pd
import numpy as np

# Lookup classes
# Each key column requires an input class which maps from the source datatype to an integer

class IntegerLookup:
    def get(self, values):
        return values

class IntegerLookupSafe:
    """An integer lookup that checks bounds"""
    def __init__(self, lower, upper):
        self.lower = lower
        self.upper = upper
    
    def get(self, values):
        # TODO: replace np.all with np.any for faster failing (probably doesn't improve performance)
        if not np.all((self.lower <= values) & (values <= self.upper)):
            raise KeyError(f'values are not between table minimum ({self.lower}) and maximum ({self.upper})')
        return values

class BoundIntLookup:
    def __init__(self, lower, upper):
        self.lower = int(lower)
        self.upper = int(upper)
    
    def get(self, numpy_array):
        return np.clip(numpy_array, self.lower, self.upper)

class BandLookup:
    def __init__(self, upper_bounds, labels):
        """Inputs must be sorted"""
        self.upper_bounds = np.array(upper_bounds)
        self.labels = np.array(labels)

    def get(self, numpy_array):
        """get the labels for vector"""
        indices = np.searchsorted(self.upper_bounds, numpy_array)
        return self.labels[indices]
    
    def get_max(self, numpy_array):
        """get the upper band limit"""
        indices = np.searchsorted(self.upper_bounds, numpy_array)
        return self.upper_bounds[indices]
    
    @classmethod
    def from_dataframe(cls, df:pd.DataFrame, band_column_name, integer_column_name):
        df_copy = df.sort_values(band_column_name)
        band_arr = df_copy[band_column_name].values
        band_labels = df_copy[integer_column_name].values
        return cls(band_arr, band_labels)
    
    @classmethod
    def from_excel_sheet(cls, workbook_path, sheet_name):
        """Extract from an excel sheet, there must be two columns, the band column and the integer column"""
        df = pd.read_excel(workbook_path, sheet_name=sheet_name)
        assert len(df.columns) == 2, "must have 2 columns"
        band_col_name, int_col_name = df.columns[0], df.columns[1]
        return cls.from_dataframe(df, band_col_name, int_col_name)

class IntKeyTable:
    """Vectorised Table Allowing for multiple integer keys"""
    def __init__(self, df: pd.DataFrame):
        self.key_cols = list(df.columns[:-1])
        self.value_col = df.columns[-1]

        # ensure all key_cols are integers
        for col in self.key_cols:
            df[col] = df[col].astype(int)

        self.df = df.sort_values(by=list(reversed(self.key_cols))) # TODO: don't need to store this in self, this but handy for testing.
        self.bases = [min(self.df[col]) for col in self.key_cols]
        self.ranges = [max(self.df[col]) - min(self.df[col]) + 1 for col in self.key_cols]

        # set up index scalars
        total_scalar = 1
        self.scalars = []
        for col in self.ranges:
            self.scalars.append(total_scalar)
            total_scalar *= col

        self.values = self.df[self.value_col].values

        expected_rows = np.prod(self.ranges)
        if expected_rows != len(self.values):
            raise ValueError(f'Input `df` is not rectangular, {expected_rows=} != {len(self.values)=}')

    def get_index(self, *keys):
        index = 0
        for key, base, scalar in zip(keys, self.bases, self.scalars):
            index += (key - base) * scalar
        return index
    
    def get_value(self, *keys):
        indices = self.get_index(*keys)
        return self.values[indices]
    
    def __getitem__(self, keys):
        return self.get_value(*keys)
    
class Table:
    """Table provides multi-key compatible high performance table lookup.
    """
    col_types = "int", "int_bound", "str", "band", "float"

    def __init__(self, df:pd.DataFrame, rectify=False, safe=True):
        """Initialise a table from a dataframe.
        
        parameters:
         - `rectify`: force table to be rectangular (default False)
         - `safe`: validates that integers are between bounds (default True)

        Tables should be in long format:
         - the final column containing the values to look up
         - all other columns contain keys to lookup
         - tables should be contingous, i.e. no gaps in integer keys.
         - tables should be complete if viewed as square matrixes (i.e. all combinations of keys are input).  If not, you should fill any gaps with np.nan or a suitable value.

        The type of key is determined by the suffix on the column name:
        `|int`: integers (...0, 1, 2, 3...), can start and end anywhere, but must be consecutive
        `|int_bound`: as `|int` but any values are constrained to the lowest and highest values.
        `|str': keys are interpreted as strings, e.g. 'M' and 'F'
        `|band`: key is numeric and treated as the upper bound on a lookup.
        `|float`: not currently available due to floating point equality, use int or band depending on use case.
        """
        if rectify:
            df = Table.rectify(df)

        key_cols = list(df.columns[:-1])
        df = df.sort_values(key_cols[::-1]) # sort by reverse order

        df_int_keys = df.copy() # this will have keys overridden as we work through mappers

        # prepare the mappers
        self.mappers = []
        for col in key_cols:
            col_type = col.split("|")[1] # "int", "str" etc
            if col_type == "int":
                if safe:
                    lower = df[col].min()
                    upper = df[col].max()                    
                    self.mappers.append(IntegerLookupSafe(lower=lower, upper=upper))
                else:
                    self.mappers.append(IntegerLookup())
            elif col_type == "int_bound":
                # bound integer forces values to be between the lowest and highest value in the table, for example maximum durations in mortality tables.
                lower = df[col].min()
                upper = df[col].max()
                self.mappers.append(BoundIntLookup(lower=lower, upper=upper))
            elif col_type in ["str", "band"]:
                df_col = pd.DataFrame(df[col].unique(), columns=["band_name"]).reset_index().sort_values("band_name")

                # add a nan on the end so we get errors if the lookup fails
                # as by default the BandLookup will return last item if no earlier matches
                # commented out as it converts the datatype - we need these to be integers
                # df_col.loc[len(df_col)] = len(df_col), np.nan
                band_mapper = BandLookup.from_dataframe(df_col, "band_name", "index")
                self.mappers.append(band_mapper)
                df_int_keys[col] = band_mapper.get(df_int_keys[col])
            else:
                raise NotImplementedError(f"{col_type} not implemented on {col}")
        
        # create an intkeytable

        self._int_key_table = IntKeyTable(df_int_keys)
        self.df = df # store for referencing TODO: decide if this should be removed.
    
    def get(self, *keys):
        assert len(keys) == len(self.mappers)
        # TODO: if just one key then this doesn't work?  (needs fixed throughout)
        int_keys = [mapper.get(key) for key, mapper in zip(keys, self.mappers)]
        return self._int_key_table.get_value(*int_keys)

    def __getitem__(self, keys):
        # print(keys, type(keys))
        if not isinstance(keys, tuple):
            keys = keys, #force to be a tuple
        return self.get(*keys)

    def __repr__(self):
        # TODO: return a nice representation of the table, e.g. head/keys etc.
        return self.df.__repr__()
    
    @staticmethod
    def rectify(df: pd.DataFrame, fill=np.nan) -> pd.DataFrame:
        """Convert a triangular (incomplete) dataframe into a valid rectangular dataframe
        
        any missing points will be filled with `fill`, default: np.nan
        """
        key_cols = list(df.columns[:-1])
        val_col = df.columns[-1]

        df_unique_keys = []
        for col_name in key_cols:
            if col_name.endswith('|int') or col_name.endswith('|int_bound'):
                # make sure all integers are covered (no gaps)
                full_keys = list(range(df[col_name].min(), df[col_name].max() + 1))
            else:
                # otherwise, grab the unique values.
                full_keys = list(df[col_name].unique())
            df_unique_keys.append(full_keys)

        # construct a dataframe from the cartesian product of these
        df_rect_keys = pd.DataFrame(itertools.product(*df_unique_keys), columns=key_cols)

        # fill out the table
        df_rect = df_rect_keys.merge(right=df, how='left', on = key_cols)

        # replace gaps with the fill value.
        df_rect[val_col] = df_rect[val_col] .fillna(fill)
        return df_rect

    @classmethod
    def read_excel(cls, spreadsheet_path, sheet_name):
        """Read in a table from an excel sheet"""
        df = pd.read_excel(spreadsheet_path, sheet_name=sheet_name)
        return cls(df)
    
    @classmethod
    def read_csv(cls, csv_path):
        df = pd.read_csv(csv_path)
        return cls(df)



import warnings
import types
from inspect import signature, getmembers
import pandas as pd

class _Cache:
    """Cache provides controllable memoization for model methods"""

    def __init__(self, func, param_len, param_names):
        self.func = func
        self.param_len = param_len
        self.param_names = param_names
        self.has_one_param = self.param_len == 1
        self._store = dict()
        self.__name__ = func.__name__

    def __call__(self, *arg):
        if arg in self._store:
            return self._store[arg]
        else:
            result = self.func(*arg)
            self._store[arg] = result
            return result

    def __repr__(self):
        return f"<Cache Function: {self.func.__name__} Params: {self.param_names} Size: {len(self._store)}>"
    
    def sum(self):
        """return the sum of all values in the Cache Function"""
        return sum(self._store.values())
    
    @property
    def keys(self):
        return list(self._store.keys())

    @property
    def values(self):
        return list(self._store.values())
    
    @property
    def df(self):
        """return the cache as a pandas dataframe"""
        df = pd.DataFrame(data=self.keys, columns=self.param_names)
        df[self.__name__] = self.values
        # df = df.set_index(list(self.param_names))  # TODO: decide if keys should be indexes or not.
        return df

class Model:
    def __init__(self, *, do_run = None, proj_len:int = 0, **kwargs,):
        """Base Class to subclass for user models.

        When the model is instanced it is run to proj_len, if non-zero.

        Parameters
        ----------
        - proj_len: length of projection to run
        

        All variables/methods in user models should be lower case, using underscore as spaces.

        Class level methods:
          RunModel(proj_len):
            Run the model to proj_len.

        Special user methods:
          BeforeRun(self):
            If this is specified in the user model it called before the projection starts, e.g. to set up some specific variables

          AfterRun(self):
            user method, called after Run is completed, e.g. can use to calculate NPVs of variables

        methods/variables to avoid:
        methods/variables starting with an underscore `_` are treated as internal.  You may break functionality if you create your own.

        """

        if do_run is not None:
            warnings.warn("Warning: `do_run` will be removed in a future version, use `proj_len` to control projection")
        else:
            do_run = proj_len > 0

        self._cached = False
        self._is_run = False
        
        if do_run:
            # TODO: this can be cleaned up once do_run is removed
            if not isinstance(proj_len, int):
                raise ValueError("proj_len must be an integer")
            elif proj_len < 0:
                raise ValueError("proj_len must be non-negative")
            self.proj_len = proj_len

        # store all keyword arguments as attributes
        for k, v in kwargs.items():
            if k in dir(self):
                # TODO: include a strict mode that raises on duplicates?
                warnings.warn("Warning: Duplicate Item: "+str(k))
            setattr(self, k, v)

        # cacheify
        self._cache_funcs()

        if do_run and proj_len > 0:
            self.RunModel(proj_len)

    def RunModel(self, proj_len: int):
        if self._is_run:
            # TODO: replace this with ability to run further, but warn that earlier values not recalculated?
            raise ValueError("Run has already been completed.")
        
        if hasattr(self, "BeforeRun"):
            self.BeforeRun()

        if not self._cached:
            raise ValueError("Functions have not been cached")  # NB: this shouldn't occur as now caching in instance
        for t in range(proj_len):
            for name, func in self._funcs.items():
                if func.has_one_param and func.param_names[0] == 't':   # skip functions with more than one parameter
                    func(t)   #call each function in turn, starting from t==0
        self._is_run = True

        if hasattr(self, "AfterRun"):
            return self.AfterRun()
    
    def _cache_funcs(self):
        if self._cached:
            raise ValueError("Cache has already been set-up, please create a new instance")

        self._funcs = {}

        for method_name, method in getmembers(self):
            if method_name[0] != "_" and method_name[0].islower() and isinstance(method, types.MethodType):
                param_count = len(signature(method).parameters) # count the parameters in the function.
                param_names = tuple(signature(method).parameters)
                cached_method = _Cache(method, param_count, param_names)
                setattr(self, method_name, cached_method)
                self._funcs[method_name] = cached_method
       
        self._cached = True
    
    def _info(self):
        """Print info about the model"""
        for name, func in self._funcs.items():
            print(f"{name}: {func}")
        
    def ToDataFrame(self, param = 't'):
        """return a pandas dataframe of all single parameter columns"""
        df = pd.DataFrame()
        for func in self._funcs:
            if self._funcs[func].has_one_param and self._funcs[func].param_names[0] == param:
                df[func] = pd.Series(self._funcs[func].values)

        # if t is in the dataframe, move it to first position
        if "t" in df.columns:
            df.insert(0, "t", df.pop("t"))
        return df
    
    @property
    def df(self):
        return self.ToDataFrame()


from .heavylight import Model
from .heavytables import Table
from .memory_optimized_model import LightModel, agg, default_agg_function
from .memory_optimized_cache import CacheGraph
from .make_examples import make_example

__version__ = '1.0.6'



from heavylight import Model

class SimpleModel(Model):
    def t(self, t):
        return t

    def num_pols_if(self, t):
        if t == 0:
            return 1
        else:
            return self.num_pols_if(t - 1) * 0.98

    def cashflow(self, t):
        return self.num_pols_if(t) * 100
    
    def v(self, t):
        """discount factor for time t --> time 0"""
        if t == 0:
            return 1
        else:
            return self.v(t - 1) / (1 + self.forward_rate(t))
    
    def forward_rate(self, t):
        return 0.04
    
    def pv_cashflow(self, t):
        """present value of the cashflow occuring at time t"""
        return self.cashflow(t) * self.v(t)

def test_heavylight():
    model = SimpleModel(do_run = True, proj_len = 10)
    assert model.pv_cashflow.sum() > 0

class UnitModel(Model):
    def t(self, t):
        return t

    def func_t(self, t):
        return 100 * t
    
    def func_a(self, a):
        return 100 * a

def test_UnitModel():
    proj_len = 10
    model = UnitModel(proj_len = proj_len)
    assert model.func_t.sum() == sum(100 * t for t in range(proj_len))
    assert model.func_a.values == []
    assert model.t.values == list(range(proj_len))
    model.func_a(5)
    model.func_a(10)
    model.func_a(7)
    assert model.func_a.values == [500, 1000, 700]  # values are ordered by insertion.
    assert [k[0] for k in model.t.keys] == model.t.values  # keys are stored as tuples, need to grab the first one.

class DataInput(Model):
    b = 4
    def t(self, t):
        return t
    def a_times_t(self, t):
        return self.a * self.b * t
    
def test_data_input():
    model = DataInput(proj_len=10, a=100)
    assert model.a_times_t(0) == 0
    assert model.a_times_t(10) == 10 * 100 * 4

class RecursiveModel(Model):
    """recursive model, including data inputs"""
    start_time = 14
    init_delta: int # = 2
    def t(self, t):
        return t

    def time_left(self, t):
        if t == 0:
            return self.start_time
        else:
            new_time = self.time_left(t - 1) - self.time_delta(t)
            return max(new_time, 0)
    
    def time_delta(self, t):
        return self.init_delta
        
def test_recursive():
    model = RecursiveModel(proj_len = 7, init_delta = 3)  # note overriding init_delta, this will warn
    assert model.time_left.values == [14, 11, 8, 5, 2, 0, 0]

class MultipleParamModel(Model):
    def t(self, t):
        return t
    
    def func_2_params(self, t, dur):
        """function with 2 parameters"""
        return t * dur * self.param_a(5)
    
    def sum_f2(self, t):
        return sum(self.func_2_params(t, dur) for dur in range(t + 1))
    
    def square_f2(self, t):
        return self.func_2_params(t, t)
    
    def param_a(self, a):
        return 3
    
def test_multi_param():
    model = MultipleParamModel(proj_len = 5)
    assert model.sum_f2(3) == 3 * 3 *(0 + 1 + 2 + 3)
    assert model.func_2_params(2, 5) == 2 * 5 * 3

class DataFrameTest(Model):
    def t(self, t):
        return t
    
    def ones(self, t):
        return 1
    
    def excluded(self, blah):
        """not included in the dataframe as t is not a parameter"""
        return 99
    
    def ninety_nines(self, t):
        return self.excluded(t)

def test_dataframe_extracts():
    model = DataFrameTest(proj_len=5)
    df = model.df
    assert len(df) == 5
    assert df['ones'].sum() == 5 
    assert df['ninety_nines'].sum() == 5 * 99
    assert sorted(df.columns) == sorted(['t', 'ones', 'ninety_nines'])

    # check dataframe for one function
    df_ones = model.ones.df
    assert df_ones.shape == (5, 2)  # 5 rows, 2 columns (t and ones)
    assert sorted(df_ones.columns) == sorted(['t', 'ones'])



# %%

import heavylight

class Annuity(heavylight.Model):
    def t(self, t):
        return t

    def expected_claim(self, t):
        return self.number_alive(t) * self.data["annuity_per_period"]

    def number_alive(self, t):
        if t == 0:
            return self.data["initial_policies"]
        else:
            return self.number_alive(t - 1) - self.deaths(t - 1)
    
    def deaths(self, t):
        return self.number_alive(t) * self.mortality_rate(t)

    def mortality_rate(self, t):
        return 0.02

    def v(self, t):
        """discount factor from time t to time 0"""
        if t == 0:
            return 1
        else:
            return self.v(t - 1) / (1 + self.forward_rate(t))
    
    def forward_rate(self, t):
        return 0.04

    def pv_expected_claim(self, t):
        return self.expected_claim(t) * self.v(t)
    
# %%

policy_data = {
    "initial_policies": 10,
    "annuity_per_period": 55,
}

# %%

model = Annuity(data = policy_data,
                do_run = True,
                proj_len = 20,
                )

print(model.pv_expected_claim.sum())

model_cashflows = model.df
print(model_cashflows)


# %%

from heavylight import Model

# %%

class SimpleModel(Model):
    def t(self, t):
        return t

    def num_pols_if(self, t):
        if t == 0:
            return 1
        else:
            return self.num_pols_if(t - 1) * 0.98

    def cashflow(self, t):
        return self.num_pols_if(t) * 100
    
    def v(self, t):
        """discount factor for time t --> time 0"""
        if t == 0:
            return 1
        else:
            return self.v(t - 1) / (1 + self.forward_rate(t))
    
    def forward_rate(self, t):
        return 0.04
    
    def pv_cashflow(self, t):
        """present value of the cashflow occuring at time t"""
        return self.cashflow(t) * self.v(t)
    
# %%

if __name__ == "__main__":
    model = SimpleModel(proj_len = 10)
    print(model.pv_cashflow.sum())

    print(model.v.values)

    print(model.df)

    print(model.pv_cashflow.df)


# -*- coding: utf-8 -*-
# required libraries:
# heavylight
# pip install heavylight

# %%
import heavylight
from heavylight import Table
from protection_model import TermAssurance

def solve_prot_premium(model: heavylight.Model, data: dict, basis: dict):
    
    data = data.copy()

    pricing_entries = {
            "annual_premium":1,
            "init_pols_if":1,
            }
    
    data.update(pricing_entries)

    proj_len = data["term_y"] * 12 + 1

    model_inst = model(data=data, basis=basis, proj_len=proj_len)
            
    # extract npvs
    npv_claims = model_inst.pv_claims.sum()
    npv_expenses = model_inst.pv_expenses.sum()
    npv_premiums = model_inst.pv_premiums.sum()
            
    # calculate premium
    
    annual_risk_premium = (npv_claims + npv_expenses) / npv_premiums
    monthly_risk_premium = annual_risk_premium / 12
    return round(monthly_risk_premium, 2)

# %%
if __name__=='__main__':
    
    basis = {
        "cost_inflation_pa": 0.02,
        "initial_expense": 500,
        "expense_pp": 10,
        "lapse_rate_pa": 0.1,
        "mort_table": Table.read_csv(r"tables/q_x_generic.csv"),
        "forward_rates": Table.read_csv(r"tables/forward_rates.csv"),
    }

    data = {
        "sum_assured": 100_000,
        "age_at_entry": 49,
        "term_y": 35,
        "smoker_status": "N",
        "shape": "level",
        "annual_premium": 1,
        "init_pols_if": 1,
        "extra_mortality": 0,
        "sex": "F"
    }

    model = TermAssurance(basis=basis, data=data, proj_len=data["term_y"]*12 + 12)
  
    monthly_premium = solve_prot_premium(TermAssurance, data, basis)
    print("Premium: ", monthly_premium)
   

    



from typing import Callable
from collections import defaultdict
from dataclasses import dataclass
from functools import wraps
from typing import Any, Callable, Tuple, Union, FrozenSet
import pandas as pd

@dataclass(frozen=True)
class FunctionCall:
    func_name: str
    args: tuple
    kwargs: FrozenSet[Tuple[str, Any]]

    def __repr__(self):
        if len(self.kwargs) == 0:
            if len(self.args) == 1:
                return f"{self.func_name}({self.args[0]})"
            return f"{self.func_name}{self.args}"
        return f"{self.func_name}({', '.join(map(str, self.args))}, {', '.join(f'{k}={v}' for k, v in self.kwargs)})"

ArgsHash = Tuple[Tuple, frozenset]

class CacheGraph:
    """
    The cache graph maintains data structures necessary for caching and memory-optimizing collections of recursive functions.
    It is applied to functions as a decorator
    """
    def __init__(self):
        self.reset()

    def reset(self):
        """
        Clear all internal state of the cache graph.
        """
        self.stack: list[FunctionCall] = [] # what function is currently being called
        self._caches: defaultdict[str, dict[ArgsHash, Any]] = defaultdict(dict) # Results of function calls, ugly keys like ((1, 2), frozenset([('a', 1)]))
        self._caches_agg: defaultdict[str, dict[ArgsHash, Any]] = defaultdict(dict)
        self.graph: defaultdict[FunctionCall, set[FunctionCall]] = defaultdict(set) # Call graph, graph[caller] = [callee1, callee2, ...]
        # Typically aggregated results for a function at a timestep.
        # What is the last function that needs the result of a function? Used to help in clearing the cache
        self.last_needed_by: dict[FunctionCall, FunctionCall] = {}
        # can_clear[caller] = [callee1, callee2, ...] means that caller can clear the cache of callee1 and callee2
        self.can_clear: dict[FunctionCall, list[FunctionCall]] = defaultdict(list)
        self.all_calls: set[FunctionCall] = set()
        self.cache_misses: defaultdict[FunctionCall, int] = defaultdict(int)

    def check_if_cached(self, function_call: FunctionCall):
        name_in_cache = function_call.func_name in self._caches
        return name_in_cache and (function_call.args, function_call.kwargs) in self._caches[function_call.func_name]
    
    def optimize(self):
        self.can_clear = defaultdict(list)
        for callee, caller in self.last_needed_by.items():
            self.can_clear[caller].append(callee)
        uncleared_calls = self.all_calls - set(self.last_needed_by.keys())
        for call in uncleared_calls:
            self.can_clear[call].append(call)

    def optimize_and_reset(self):
        self.optimize()
        can_clear = self.can_clear
        self.reset()
        self.can_clear = can_clear

    def __call__(self, storage_func: Union[Callable[[int], Any], None] = None):
        def custom_cache_decorator(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                frozen_kwargs = frozenset(kwargs.items())
                function_call = FunctionCall(func.__name__, args, frozen_kwargs)
                if self.stack:
                    self.graph[self.stack[-1]].add(function_call)
                    self.last_needed_by[function_call] = self.stack[-1]
                if not self.check_if_cached(function_call):
                    self.all_calls.add(function_call)
                    self.cache_misses[function_call] += 1
                    self.stack.append(function_call)
                    result = func(*args, **kwargs)
                    self._caches[func.__name__][(args, frozen_kwargs)] = result
                    for clearable_call in self.can_clear[function_call]:
                        del self._caches[clearable_call.func_name][(clearable_call.args, clearable_call.kwargs)]
                    self.stack.pop()
                    self._store_result(storage_func, func, (args, frozen_kwargs), result)
                    return result
                return self._caches[func.__name__][(args, frozen_kwargs)]
            decorator = CacheMethod(self, wrapper, storage_func)
            return decorator
        return custom_cache_decorator
    
    def _store_result(self, storage_func: Union[Callable, None], func: Callable, args_hash: ArgsHash, raw_result: Any):
        """We might want to store an intermediate result"""
        if storage_func is None:
            return
        stored_result = storage_func(raw_result)
        self._caches_agg[func.__name__][args_hash] = stored_result

    def size(self):
        return sum(len(cache) for cache in self._caches.values())
    
    @property
    def cache(self):
        caches = defaultdict(dict, {func_name: {get_pretty_key(k): v for k, v in cache.items()} for func_name, cache in self._caches.items()})
        return caches
    
    @property
    def cache_agg(self):
        caches = defaultdict(dict, {func_name: {get_pretty_key(k): v for k, v in cache.items()} for func_name, cache in self._caches_agg.items()})
        return caches

class CacheMethod:
    def __init__(self, cache_graph: CacheGraph, func: Callable, agg_func: Union[Callable, None] = None):
        self.cache_graph = cache_graph
        self.func = func
        self.agg_func = agg_func

    @property
    def df(self):
        return pd.DataFrame({self.func.__name__: self.cache})

    @property
    def df_agg(self):
        return pd.DataFrame({self.func.__name__: self.cache_agg})

    # only run the dictionary comprehension for the particular method we want to access
    # simply returning self.cache_graph.caches[self.func.__name__] would run the dictionary comprehension for all methods
    @property
    def cache(self):
        return {get_pretty_key(k): v for k, v in self._cache.items()}
    
    @property
    def cache_agg(self):
        return {get_pretty_key(k): v for k, v in self._cache_agg.items()}

    @property
    def _cache(self):
        return self.cache_graph._caches[self.func.__name__]

    @property
    def _cache_agg(self):
        return self.cache_graph._caches_agg[self.func.__name__]

    def __setitem__(self, key, value):
        if isinstance(key, int):
            self._cache[((key,), frozenset())] = value
        else:
            self._cache[(key, frozenset())] = value

    def __repr__(self):
        return f"<Cache Function: {self.func.__name__}, Size: {len(self._cache)}>"
    
    def __call__(self, *args: Any, **kwds: Any) -> Any:
        return self.func(*args, **kwds)
    
def get_pretty_key(key: ArgsHash):
    if len(key[0]) == 1 and len(key[1]) == 0:
        return key[0][0]
    elif len(key[1]) == 0:
        return key[0]
    else:
        return key
    


import numpy as np
import pandas as pd
import heavylight.heavytables as ht

def test_integer_lookup():
    lookup = ht.IntegerLookup()
    arr = np.array([1,2,3])
    assert np.all(arr==lookup.get(arr))

def test_bound_integer_lookup():
    lookup = ht.BoundIntLookup(0, 10)
    arr = np.array([-1, 0, 1, 10, 11])
    assert np.all(np.array([0, 0, 1, 10, 10])==lookup.get(arr))

def test_rectify():
    df1 = pd.DataFrame({'a': [1, 2], 'b': [1, 2], 'c': [1, 2]})
    expected_rectified_df1 = pd.DataFrame({'a': [1, 1, 2, 2], 'b': [1, 2, 1, 2], 'c': [1, np.nan, np.nan, 2]})
    assert ht.Table.rectify(df1).equals(expected_rectified_df1)


from heavylight.memory_optimized_cache import FunctionCall, CacheGraph
import pytest

def test_function_call():
    fc_args_kwargs = FunctionCall("func", (1, 2), frozenset([('a', 1)]))
    assert fc_args_kwargs.func_name == "func"
    assert fc_args_kwargs.args == (1, 2)
    assert fc_args_kwargs.kwargs == frozenset([('a', 1)])
    assert repr(fc_args_kwargs) == "func(1, 2, a=1)"
    fc_single_arg_no_kwargs = FunctionCall("func", (1,), frozenset())
    assert repr(fc_single_arg_no_kwargs) == "func(1)"
    fc_multiple_args_no_kwargs = FunctionCall("func", (1, "hello"), frozenset())
    assert repr(fc_multiple_args_no_kwargs) == "func(1, 'hello')"

def test_cache_graph_storage_function():
    cg = CacheGraph()
    @cg(lambda x: x**2)
    def fib(n):
        if n < 2:
            return n
        return fib(n - 1) + fib(n - 2)
    fib(5)
    assert len(cg.cache_agg["fib"]) == 6
    for k, v in cg.cache_agg["fib"].items():
        assert v == fib(k)**2

def test_cache_dunders():
    cg = CacheGraph()
    @cg()
    def fib(n):
        if n < 2:
            return n
        return fib(n - 1) + fib(n - 2)
    fib(5)
    assert repr(fib) == "<Cache Function: fib, Size: 6>"
    assert len(fib.cache) == 6
    test_key = 5
    assert fib.cache[test_key] == cg.cache['fib'][test_key] == 5
    assert fib.cache[5] == 5 # prettified keys
    fib[5] = 10
    assert fib.cache[test_key] == cg.cache['fib'][test_key] == 10
    assert fib.cache[5] == 10
    fib[(5,)] = 100
    assert fib.cache[test_key] == cg.cache['fib'][test_key] == 100

@pytest.mark.timeout(4)
def test_cache_graph_memory_optimization():
    cg = CacheGraph()

    @cg()
    def fib(n: int):
        if n <= 1:
            return n
        return fib(n-1) + fib(n-2)

    def get_max_memory_fib(n: int, cg: CacheGraph):
        max_memory_size = 0
        for i in range(n+1):
            fib(i)
            max_memory_size = max(max_memory_size, cg.size())
        return max_memory_size

    assert get_max_memory_fib(200, cg) == 201
    cg.optimize_and_reset()
    assert get_max_memory_fib(200, cg) == 2